/*******************************************************************************
* manager.cpp:                                                                 *
* (c)1999 VideoLAN                                                             *
*------------------------------------------------------------------------------*
*                                                                              *
*******************************************************************************/


//------------------------------------------------------------------------------
// Preamble
//------------------------------------------------------------------------------
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "config.h"
#include "common.h"
#include "string.h"
#include "log.h"
#include "admin.h"
#include "cfgfile.h"

#include "debug.h"


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
C_Admin::C_Admin(handle hLogger)
{
  hLog = hLogger;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
C_Admin::~C_Admin()
{
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
int C_Admin::Init(const C_String& strCfgFile)
{
  C_CfgFile cCfgFile(hLog);
  C_String strPort, strHost;
  
  // Open the cfgfile
  int iRc = cCfgFile.Open(strCfgFile);

  // Find the section where is stored the config
  if (!iRc)
    iRc = cCfgFile.FindSection("Admin");
  if (iRc)
    Log(hLog, LOG_ERROR, "Init failed: no 'Admin' section found in cfg file");
  
  // Parse the file
  C_String strVarName, strVarVal;
  while(!iRc)
  {
    // Read the current entry
    iRc = cCfgFile.Read(&strVarName, &strVarVal);
    
    // Check it is OK
    if (iRc == CFG_OK)
    {
      // 
      if(strVarName == "Port")
      {
        strPort = strVarVal;
      }
      else if(strVarName == "Host")
      {
        strHost = strVarVal;

        // Debug only ???
        Log(hLog, LOG_NOTE, "Host parameter in section Admin ignored");
      }
    }
  }

  if(iRc == CFG_END)
    iRc = cCfgFile.Close();

  // Wait for incoming connections
  if(!iRc)
    iRc = OpenSocket(strPort);

  if(!iRc)
    Log(hLog, LOG_NOTE, "Admin initialised");
  else
    Log(hLog, LOG_ERROR, "Admin init failed");

  return iRc;
}


//------------------------------------------------------------------------------
// Process incoming connections
//------------------------------------------------------------------------------
// TO DO: error processing
//------------------------------------------------------------------------------
int C_Admin::Run()
{
  int iNewSockFd;
  struct sockaddr_in sRemoteAddr;
  socklen_t iRemoteAddrLen = sizeof(sRemoteAddr);
  int iRc = 0;

  Log(hLog, LOG_NOTE, "Waiting for incoming administrative connections");

  for (;;)
  { 
    // Create a socket to deal with that connection. The socket will be the
    // property of the administrator thread 
    iNewSockFd = accept(iSockFd, (struct sockaddr *)&sRemoteAddr, &iRemoteAddrLen);

    if (iNewSockFd < 0)
    {
      // An error occured */
      Log(hLog, LOG_ERROR, "Unable to process incoming connection from %s: %s",
          inet_ntoa(sRemoteAddr.sin_addr), strerror(errno));

      // It may be a fast loop, so slow it down
      usleep(10000);

      // Retry with another connection
      continue;
    }
    else
    { 
      Log(hLog, LOG_NOTE, "Processing incoming connection from %s...",
          inet_ntoa(sRemoteAddr.sin_addr));

      iRc = ProcessConnection(iNewSockFd);
    }
  }
  
  return 0;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
int C_Admin::Destroy()
{
  int iRc = close(iSockFd);
  if(iRc)
    Log(hLog, LOG_WARN, "Could close socket: %s", strerror(errno));

  return iRc;
}



//------------------------------------------------------------------------------
// Open and initialize the socket that will wait for administrative connections
//------------------------------------------------------------------------------
int C_Admin::OpenSocket(const C_String& strPort)
{
  int iRc = NO_ERR;

  // Open socket for waiting on TCP connections */
  iSockFd = socket (AF_INET, SOCK_STREAM, 0);
  
  if(iSockFd < 0)
  {
    Log(hLog, LOG_ERROR, "Cannot open a socket: %s", strerror(errno));
    iRc = NET_ERR;
  }

  // Init the socket descriptor */
  struct sockaddr_in sServerAddr;
  socklen_t iAddrLen = sizeof(sServerAddr);
  if(!iRc)
  {
    memset(&sServerAddr, 0, sizeof(sServerAddr));
    sServerAddr.sin_family = AF_INET;

    // Receive on all local addresses
    sServerAddr.sin_addr.s_addr = htonl(INADDR_ANY);

    // Set the port we want to listen to
    int iPort = atoi(strPort);
    if (iPort)
      sServerAddr.sin_port = htons(iPort);
    else
    {
      // It's given by its service name
      struct servent* sp;
      if ((sp = getservbyname(strPort, "tcp")) == NULL)
      {
	iRc = errno;
	Log(hLog, LOG_ERROR, "Unknown service port: %s", (const char*)strPort);
      }
      else
	sServerAddr.sin_port = sp->s_port;
    }
  }

  // Bind the socket
  if (!iRc)
  {
    iRc = bind(iSockFd, (struct sockaddr *)&sServerAddr,
               sizeof(struct sockaddr));
    
    if(iRc)
      Log(hLog, LOG_ERROR, "Unable to bind the socket: %s", strerror (iRc));
  }

  // Listen for incoming connections
  if (!iRc)
  {
    iRc = listen(iSockFd, MAX_ADMIN_CONN);
    
    if(iRc)
      Log(hLog, LOG_ERROR, "Unable to listen to incoming connections: %s",
          strerror(iRc));
    else
    {
      getsockname(iSockFd, (struct sockaddr*)&sServerAddr, &iAddrLen);
      Log(hLog, LOG_NOTE, "Listening on address %s port %d",
          inet_ntoa(sServerAddr.sin_addr), ntohs(sServerAddr.sin_port));
    }
  }

  return iRc;
}


//------------------------------------------------------------------------------
// Wait for and manage incoming connections
//------------------------------------------------------------------------------
int C_Admin::ProcessConnection(int iSockFd)
{
  send(iSockFd, "Coucou\n", 8, 0);
  close(iSockFd);
  return 0;
}
