/*******************************************************************************
* input.cpp: Input streams                                                     *
* (c)1999 VideoLAN                                                             *
*------------------------------------------------------------------------------*
* Hardware management                                                          *
*******************************************************************************/


//------------------------------------------------------------------------------
// Preamble
//------------------------------------------------------------------------------
#include <pthread.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>

#include "config.h"
#include "common.h"
#include "string.h"
#include "log.h"
#include "cfgfile.h"
#include "list.h"
#include "directory.h"
#include "ts.h"
#include "buffer.h"
#include "output.h"
#include "input.h"

#include "debug.h"
#include "list.cpp"



//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
C_Input::C_Input()
{
  pDirectory = NULL;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
C_Input::~C_Input()
{
  if(pDirectory)
    delete pDirectory;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
// The Log of the manager is used, that's why a hManagerLog handle is passed to
// the method
//------------------------------------------------------------------------------
int C_Input::Init(const C_String& strCfgFile, const C_String& strSection)
{
  // First register the source as a log client
  int iRc = NO_ERR;
  hLog = pLog->Register((const char*)strSection, LOG_DBGMSG|LOG_FILE|LOG_SCR);
  if (!hLog)
    iRc = GEN_ERR;

  // Then create the directory
  if(!iRc)
  {
    pDirectory = new C_Directory(hLog);
    ASSERT(pDirectory);
  }
  
  // Now call the child's preinitialisation method
  if(!iRc)
    iRc = PreInit();

  // Then init the input according to the parameters stored in the cfg file
  // To start with, open that file
  C_CfgFile CfgFile(hLog);
  if(!iRc)
    iRc = CfgFile.Open(strCfgFile);
  
  // Find the section where is stored the config
  if (!iRc)
    iRc = CfgFile.FindSection(strSection);

  // Parse the file
  C_String strVarName, strVarVal;
  while (!iRc)
  {
    iRc = CfgFile.Read(&strVarName, &strVarVal);    
    if (iRc == CFG_OK)
    {
      // Call the child defined method to handle the parameter
      iRc = InitParam(strVarName, strVarVal);
    }
  }

  // Don't need the file anymore
  if(iRc == CFG_END)
    iRc = CfgFile.Open(strCfgFile);

  // Then ends the initialisation with the child's PostInit() method
  if (!iRc)
    iRc = PostInit();
  
  // Now the input is completely initialised: we can build the program table to
  // finish with the init praocess
  if(!iRc)
    iRc = BuildPgrmTable();

  return iRc;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
int C_Input::Destroy()
{
  int iRc = NO_ERR;
  
  // Stop and destroy all the outputs that are running
  for(unsigned int iIndex = 0; iIndex < cOutputList.Size(); iIndex++)
  {
    iRc |= cOutputList[iIndex].Destroy();
  }

  // Do some child specific cleanings
  iRc |= Clean();

  // Unregister the input by the logger
  iRc |= pLog->Unregister(hLog);

  return iRc;
}



//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
// At that time, only C_NetOutput are possible
//------------------------------------------------------------------------------
int C_Input::StartStreaming(unsigned int iIndex, const C_OutputCfg& cOutputCfg)
{
  C_Output* pOutput = NULL;
  int iRc = NO_ERR;
  
  // TO DO: Verifier que pgrm est bien dans la liste des pgrm disponibles

   switch(cOutputCfg.iType)
   {
     case OUTPUT_NET:
       pOutput = new C_NetOutput(cNetList, hLog);
       break;
     default:
       iRc = GEN_ERR;
       Log(hLog, LOG_ERROR, "Uknown output type: %d", cOutputCfg.iType);
   }
  
  if(pOutput)  
  {
    // Find the description of the pgrm
    C_Program* pProgram = pDirectory->GetPgrm(iIndex);
    ASSERT(pProgram);
    
    // Init the Output
    iRc = pOutput->Init(cOutputCfg);

    if(!iRc)
      iRc = StartReception(pOutput, pProgram);
    
    // Change the pgrm status
    if(!iRc)
      pProgram->SetStatus(PGRM_BROADCASTED);
    
    // Register the output or delete it if there was a pb
    if(!iRc)
      iRc = Register(pOutput);
    else
    {
      pOutput->Destroy();
      delete pOutput;
    }
  }

  return iRc;
}



//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
// At that time, only C_NetOutput are possible
//------------------------------------------------------------------------------
int C_Input::StopStreaming(unsigned int iIndex)
{
  int iRc = NO_ERR;
  
  // Find the description of the program
  C_Program* pProgram = pDirectory->GetPgrm(iIndex);
  ASSERT(pProgram);

  // Ask the child to stop the streaming
  if(pProgram->GetStatus() == PGRM_BROADCASTED)
  {
    iRc = StopReception(pProgram);
    if(!iRc)
      pProgram->SetStatus(PGRM_AVAILABLE);
  }
  else
  {
    Log(hLog, LOG_WARN, "Program not broadcasted");
    iRc = GEN_ERR;
  }

  // Unregister the output and delete it
  // TO DO

  return iRc;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
int C_Input::Register(C_Output* pOutput)
{
  ASSERT(pOutput);

  int iRc = cOutputList.PushEnd(pOutput);  
  return iRc;
}
