/*******************************************************************************
* output.cpp: output streams                                                   *
* (c)1999 VideoLAN                                                             *
*------------------------------------------------------------------------------*
*                                                                              *
*******************************************************************************/


//------------------------------------------------------------------------------
// Preamble
//------------------------------------------------------------------------------
#include <pthread.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "config.h"
#include "common.h"
#include "string.h"
#include "log.h"
#include "list.h"
#include "directory.h"
#include "ts.h"
#include "buffer.h"
#include "output.h"

#include "debug.h"


//
// Local definitions
//
#define TS_IN_ETHER 7                // TS packets that fit in an ethernet frame




//******************************************************************************
//
//******************************************************************************
//
//******************************************************************************

//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
C_OutputCfg::C_OutputCfg(byte iOutputType)
{
  iType = iOutputType;
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
C_OutputCfg::~C_OutputCfg()
{
}





//******************************************************************************
//
//******************************************************************************
//
//******************************************************************************

//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
C_Output::C_Output(C_NetList& cNetList, handle hLogger, int iBuffSize)
{
  // Create the buffer used to store incoming TS packet before sending them
  pBuff = new C_Fifo(iBuffSize);

  // NetList
  pNetList = &cNetList;

  // Logging
  hLog = hLogger;
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
C_Output::~C_Output()
{
  delete pBuff;
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
// Bla Bla sur buffer
//------------------------------------------------------------------------------
int C_Output::Send(C_TsPacket* pPacket)
{
  ASSERT(pPacket);
  int iRc = 0;

  if(pBuff->Stored() == pBuff->Size())
  {
    // Send the buffered packets to the device
    iRc = WriteToPort();
  }

  // Append the packet to the buffer or trash it
  if(!iRc)
    iRc = pBuff->Push(pPacket);
  else
    pNetList->ReleasePacket(pPacket);

  return iRc;
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
int C_Output::Flush()
{
  return WriteToPort();
}



//******************************************************************************
//
//******************************************************************************
//
//******************************************************************************


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
C_NetOutput::C_NetOutput(C_NetList& cNetList, handle hLogger) :
                C_Output(cNetList, hLogger, TS_IN_ETHER)
{
  pBuffVector = new struct iovec[TS_IN_ETHER];
  ASSERT(pBuffVector);

  // Init the array of iovec
  for(int iIndex = 0; iIndex < TS_IN_ETHER; iIndex++)
  {
    pBuffVector[iIndex].iov_base = NULL;
    pBuffVector[iIndex].iov_len = TS_PACKET_LEN;
  }
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
C_NetOutput::~C_NetOutput()
{
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
int C_NetOutput::Init(const C_OutputCfg& cCfg)
{
  int iRc = NO_ERR;
  socklen_t iSrcLen, iDstLen;
  
  // Build source address
  iRc = BuildInetAddr(&saSrc, cCfg.strSrcHost, cCfg.strSrcPort);
  iSrcLen = sizeof(saSrc);

  // Build destination address
  if (!iRc)
  {
    iRc = BuildInetAddr(&saDst, cCfg.strDstHost, cCfg.strDstPort);
    iDstLen = sizeof(saDst);
  }

  // Open a socket
  if (!iRc)
  {
    iSockFd = socket(AF_INET, SOCK_DGRAM, 0);
    if(iSockFd < 0)
    {
      Log(hLog, LOG_ERROR, "Socket error: %s\n", strerror(errno));
      iRc = NET_ERR;
    }
  }

  // Bind it to the local address
  if (!iRc)
  {
    iRc = bind(iSockFd, (struct sockaddr*)&saSrc, iSrcLen);
    if (iRc)
      Log(hLog, LOG_ERROR, "Bind error: %s\n", strerror(errno));
  }
  
  // Connect it
  if (!iRc)
  {
    iRc = connect(iSockFd, (struct sockaddr*)&saDst, iDstLen);
    if (iRc)
      Log(hLog, LOG_ERROR, "Connect error: %s\n", strerror(errno));
  }
  
  return iRc;
}


//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
int C_NetOutput::Destroy()
{
  return close(iSockFd);
}


//------------------------------------------------------------------------------
// A reecrire
//------------------------------------------------------------------------------
// If the buffer is empty when the method is called, then BUFFER_EMPTY is
// returned
//------------------------------------------------------------------------------
int C_NetOutput::WriteToPort()
{
  int iPacketNumber = pBuff->Stored();
  int iRc = NO_ERR;
  
  if(iPacketNumber == 0)
  {
    // nothing to do
    iRc = BUFFER_EMPTY;
  }
  else
  {
    // To avoid problems with the iovec array...
    ASSERT(iPacketNumber <= TS_IN_ETHER);
  
    // Fill in the iovec array
    for(int iIndex = 0; iIndex < iPacketNumber; iIndex++)
    {
      // Reference the raw data of the TS packet in the iovec array.
      // ((*pBuff)[iIndex]) returns a pointer to a TS packet. then we call the
      // const byte* () operator by the (const byte*) explicit cast to get the
      // address of the raw buffer. And finally we cast this address to (void*)
      // to avoid a warning at compilation
      pBuffVector[iIndex].iov_base = (void*)(const byte*)(*((*pBuff)[iIndex]));
//      printf("Output: %x %x %x %x\n", ((byte*)pBuffVector[iIndex].iov_base)[0], ((byte*)pBuffVector[iIndex].iov_base)[1], ((byte*)pBuffVector[iIndex].iov_base)[2], ((byte*)pBuffVector[iIndex].iov_base)[3]);
    }
  
    // Send the data that were stored in the buffer
    iRc = writev(iSockFd, pBuffVector, iPacketNumber);
    if(iRc < 0)
    {
      Log(hLog, LOG_WARN, "Send error: %s", strerror(errno));
      iRc = NET_ERR;
    }
    else
      iRc = NO_ERR;

    // Free the unused packets
    C_TsPacket* pPacket;
    for(int iIndex = 0; iIndex < iPacketNumber; iIndex++)
    {
      // Pop the packet from the buffer
      pPacket = pBuff->Pop();
      // And release it
      pNetList->ReleasePacket(pPacket);
    }
  }
  
  return iRc;
}


//------------------------------------------------------------------------------
// BuildInetAddr: build an Internet address descriptor
//------------------------------------------------------------------------------
// Build an internet address descriptor from a host name, and a port. If the
// address or the port is invalid, then GEN_ERR will be returned
//------------------------------------------------------------------------------
int C_NetOutput::BuildInetAddr(struct sockaddr_in* pInAddr,
                               const C_String& strHost, const C_String& strPort)
{
  ASSERT(pInAddr);
  int iRc = NO_ERR;
  
  // Init the structure
  memset(pInAddr, 0, sizeof(struct sockaddr_in));
  pInAddr->sin_family = AF_INET;

  // Set the port field
  int iPort = atoi(strPort);
  if (iPort)
  {
    // Port is given by its number
    pInAddr->sin_port = htons(iPort);
  }
  else
  {
    // It should be given by its service name
    struct servent* pPort = getservbyname(strPort, "udp");
    if (pPort == NULL)
    {
      Log(hLog, LOG_WARN, "Unknown port: %s/udp", (const char*)strPort);
      iRc = GEN_ERR;
    }
    else
      pInAddr->sin_port = pPort->s_port;
  } 

  // Set the address field
  if (!iRc)
  {
    // Assume the address is given in dotted-decimal format and try to convert
    if (inet_aton((const char*)strHost, &pInAddr->sin_addr) != 0)
    {  
      // Address should be given by a host name
      struct hostent* pHostEnt = gethostbyname(strHost);
      if (pHostEnt == NULL)
      {
        Log(hLog, LOG_WARN, "Unknown host: %s", (const char*)strHost);
        iRc = GEN_ERR;
      }
      else
        memcpy(&pInAddr->sin_addr, pHostEnt->h_addr, pHostEnt->h_length);
    }
  }

  return iRc;
}
